"""
# pyTDGL

Time-dependent Ginzburg-Landau in Python

![PyPI](https://img.shields.io/pypi/v/tdgl)
![GitHub Workflow Status](https://img.shields.io/github/actions/workflow/status/loganbvh/py-tdgl/lint-and-test.yml?branch=main)
[![Documentation Status](https://readthedocs.org/projects/py-tdgl/badge/?version=latest)](https://py-tdgl.readthedocs.io/en/latest/?badge=latest)
[![codecov](https://codecov.io/gh/loganbvh/py-tdgl/branch/main/graph/badge.svg?token=VXdxJKP6Ag)](https://codecov.io/gh/loganbvh/py-tdgl)
![GitHub](https://img.shields.io/github/license/loganbvh/py-tdgl)
[![Code style: black](https://img.shields.io/badge/code%20style-black-000000.svg)](https://github.com/psf/black)

## Motivation
`pyTDGL` solves a 2D generalized time-dependent Ginzburg-Landau (TDGL) equation, enabling simulations of vortex and phase dynamics in thin film superconducting devices.

## Learn `pyTDGL`

The documentation for `pyTDGL` can be found at [py-tdgl.readthedocs.io](https://py-tdgl.readthedocs.io/en/latest/).

## Try `pyTDGL`

Click the badge below and navigate to `docs/notebooks/` to try `pyTDGL` interactively online via [Binder](https://mybinder.org/)

[![Binder](https://mybinder.org/badge_logo.svg)](https://mybinder.org/v2/gh/loganbvh/py-tdgl/HEAD)

## Acknowledgments

Parts of this package have been adapted from [`SuperDetectorPy`](https://github.com/afsa/super-detector-py), a GitHub repo authored by [Mattias Jönsson](https://github.com/afsa). Both `SuperDetectorPy` and `py-tdgl` are released under the open-source MIT License. If you use either package in an academic publication or similar, please consider citing the following:

- Mattias Jönsson, Theory for superconducting few-photon detectors (Doctoral dissertation), KTH Royal Institute of Technology (2022) ([Link](http://urn.kb.se/resolve?urn=urn:nbn:se:kth:diva-312132))
- Mattias Jönsson, Robert Vedin, Samuel Gyger, James A. Sutton, Stephan Steinhauer, Val Zwiller, Mats Wallin, Jack Lidmar, Current crowding in nanoscale superconductors within the Ginzburg-Landau model, Phys. Rev. Applied 17, 064046 (2022) ([Link](https://journals.aps.org/prapplied/abstract/10.1103/PhysRevApplied.17.064046))

The user interface is adapted from [`SuperScreen`](https://github.com/loganbvh/superscreen).
"""

from setuptools import find_packages, setup

DESCRIPTION = "pyTDGL: Time-dependent Ginzburg-Landau in Python."
LONG_DESCRIPTION = __doc__

NAME = "tdgl"
AUTHOR = "Logan Bishop-Van Horn"
AUTHOR_EMAIL = "logan.bvh@gmail.com"
URL = "https://github.com/loganbvh/py-tdgl"
LICENSE = "MIT"
PYTHON_VERSION = ">=3.8, <3.11"

INSTALL_REQUIRES = [
    "cloudpickle",
    "h5py",
    "joblib",
    "jupyter",
    "matplotlib",
    "meshpy",
    "numpy",
    "pint",
    "pytest",
    "pytest-cov",
    "scipy",
    "shapely",
    "tqdm",
]

EXTRAS_REQUIRE = {
    "dev": [
        "black",
        "isort",
        "pre-commit",
    ],
    "docs": [
        "IPython",
        # https://stackoverflow.com/a/71069918
        "sphinx>=4.3.0",
        "sphinx-rtd-theme>=0.5.1",
        "sphinx-autodoc-typehints",
        "nbsphinx",
        "pillow",
        "sphinx_toolbox",
        "enum_tools",
        "sphinx-argparse",
        "sphinxcontrib-bibtex",
    ],
    "jax": [
        "jax[cpu]",
    ],
}

CLASSIFIERS = [
    "Development Status :: 3 - Alpha",
    "Intended Audience :: Science/Research",
    "License :: OSI Approved :: MIT License",
    "Operating System :: MacOS",
    "Operating System :: POSIX",
    "Operating System :: Unix",
    "Operating System :: Microsoft :: Windows",
    "Programming Language :: Python",
    "Programming Language :: Python :: 3.8",
    "Programming Language :: Python :: 3.9",
    "Programming Language :: Python :: 3.10",
    "Intended Audience :: Science/Research",
    "Topic :: Scientific/Engineering",
    "Topic :: Scientific/Engineering :: Physics",
]

PLATFORMS = ["Linux", "Mac OSX", "Unix", "Windows"]
KEYWORDS = "superconductor vortex Ginzburg-Landau"

exec(open("tdgl/version.py").read())

setup(
    name=NAME,
    version=__version__,  # noqa: F821
    author=AUTHOR,
    author_email=AUTHOR_EMAIL,
    url=URL,
    license=LICENSE,
    packages=find_packages(),
    include_package_data=True,
    description=DESCRIPTION,
    long_description=LONG_DESCRIPTION,
    long_description_content_type="text/markdown",
    keywords=KEYWORDS,
    classifiers=CLASSIFIERS,
    platforms=PLATFORMS,
    python_requires=PYTHON_VERSION,
    install_requires=INSTALL_REQUIRES,
    extras_require=EXTRAS_REQUIRE,
)



from .edge_mesh import EdgeMesh
from .mesh import Mesh
from .operators import MeshOperators
from .util import (
    compute_voronoi_polygon_areas,
    generate_voronoi_vertices,
    get_convex_polygon_area,
    get_dual_edge_lengths,
    get_edge_lengths,
    get_edges,
    get_supercurrent,
    get_voronoi_polygon_indices,
)



import os

import pytest

from . import non_gui_backend

TESTDIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "test")


def run():
    with non_gui_backend():
        pytest.main([TESTDIR])


if __name__ == "__main__":
    run()



from typing import Sequence, Tuple

import h5py
import numpy as np

from .util import get_dual_edge_lengths, get_edges


class EdgeMesh:
    """A mesh composed of the edges in a triangular mesh.

    .. tip::

        Use :meth:`EdgeMesh.from_mesh` to create from an existing mesh.

    Args:
        centers: The (x, y) coordinates for the edge_centers.
        edges: The edges as a sequence of indices.
        boundary_edge_indices: Edges on the boundary.
        directions: Directions of the edges.
        edge_lengths: Lengths of the edges.
        dual_edge_lengths: Length of the dual edges.
    """

    def __init__(
        self,
        centers: Sequence[Tuple[float, float]],
        edges: Sequence[Tuple[int, int]],
        boundary_edge_indices: Sequence[int],
        directions: Sequence[Tuple[float, float]],
        edge_lengths: Sequence[float],
        dual_edge_lengths: Sequence[float],
    ):
        self.centers = np.asarray(centers)
        self.edges = np.asarray(edges)
        self.boundary_edge_indices = np.asarray(boundary_edge_indices, dtype=np.int64)
        self.directions = np.asarray(directions)
        self.edge_lengths = np.asarray(edge_lengths)
        self.dual_edge_lengths = np.asarray(dual_edge_lengths)

    @property
    def x(self) -> np.ndarray:
        """The x-coordinates of the edge centers."""
        return self.centers[:, 0]

    @property
    def y(self) -> np.ndarray:
        """The y-coordinates of the edge centers."""
        return self.centers[:, 1]

    @staticmethod
    def from_mesh(
        sites: np.ndarray,
        elements: np.ndarray,
        dual_sites: np.ndarray,
    ) -> "EdgeMesh":
        """Create edge mesh from mesh.

        Args:
            sites: The (x, y) coordinates for the mesh vertices.
            elements: Elements for the mesh.
            dual_sites: The (xm y) coordinates for the dual mesh vertices.

        Returns:
            The edge mesh.
        """
        edges, is_boundary = get_edges(elements)
        # Get the indices of the boundary edges
        boundary_edge_indices = np.where(is_boundary)[0]
        # Shape (m, 2, 2), (sites, edges, spatial dimensions)
        edge_coords = sites[edges]
        edge_centers = edge_coords.mean(axis=1)
        directions = np.diff(edge_coords, axis=1).squeeze()
        edge_lengths = np.linalg.norm(directions, axis=1)
        dual_edge_lengths = get_dual_edge_lengths(
            edge_centers,
            elements,
            dual_sites,
            edges,
        )
        return EdgeMesh(
            edge_centers,
            edges,
            boundary_edge_indices,
            directions,
            edge_lengths,
            dual_edge_lengths,
        )

    def to_hdf5(self, h5group: h5py.Group) -> None:
        """Save the data to a HDF5 file.

        Args:
            h5group: The HDF5 group to write the data to.
        """
        h5group["centers"] = self.centers
        h5group["edges"] = self.edges
        h5group["boundary_edge_indices"] = self.boundary_edge_indices
        h5group["directions"] = self.directions
        h5group["edge_lengths"] = self.edge_lengths
        h5group["dual_edge_lengths"] = self.dual_edge_lengths

    @classmethod
    def from_hdf5(cls, h5group: h5py.Group) -> "EdgeMesh":
        """Load edge mesh from file.

        Args:
            h5group: The HDF5 group to load from.

        Returns:
            The loaded edge mesh.
        """
        if not (
            "centers" in h5group
            and "edges" in h5group
            and "boundary_edge_indices" in h5group
            and "directions" in h5group
            and "edge_lengths" in h5group
            and "dual_edge_lengths" in h5group
        ):
            raise IOError("Could not load edge mesh due to missing data.")
        return EdgeMesh(
            centers=np.array(h5group["centers"]),
            edges=np.array(h5group["edges"], dtype=np.int64),
            boundary_edge_indices=np.array(h5group["boundary_edge_indices"], np.int64),
            directions=np.array(h5group["directions"]),
            edge_lengths=np.array(h5group["edge_lengths"]),
            dual_edge_lengths=np.array(h5group["dual_edge_lengths"]),
        )



from typing import Tuple

import numpy as np

from ..em import current_loop_vector_potential
from ..parameter import Parameter


def loop_vector_potential(
    x,
    y,
    z,
    *,
    current: float,
    radius: float,
    center: Tuple[float, float, float] = (0, 0, 0),
    current_units: str = "uA",
    field_units: str = "mT",
    length_units: str = "um",
):
    if z.ndim == 0:
        z = z * np.ones_like(x)
    positions = np.array([x.squeeze(), y.squeeze(), z.squeeze()]).T
    A = current_loop_vector_potential(
        positions,
        loop_center=center,
        loop_radius=radius,
        current=current,
        current_units=current_units,
        length_units=length_units,
    )
    return A.to(f"{field_units} * {length_units}").magnitude


def CurrentLoop(
    *,
    current: float,
    radius: float,
    center: Tuple[float, float, float],
    current_units: str = "uA",
    field_units: str = "mT",
    length_units: str = "um",
) -> Parameter:
    """Returns a Parameter that computes the vector potential due to a 1D loop of current.

    Args:
        current: The current flowing in the loop.
        radius: The loop radius.
        center: The ``(x, y, z)`` position of the loop center.
        current_units: Units for current.
        field_units: Units for magnetic field.
        length_units: Units for length.

    Returns:
        A Parameter that computes the vector potential.
    """
    return Parameter(
        loop_vector_potential,
        current=current,
        radius=radius,
        center=center,
        current_units=current_units,
        field_units=field_units,
        length_units=length_units,
    )



import os
from dataclasses import dataclass
from typing import Union


class SolverOptionsError(ValueError):
    pass


@dataclass
class SolverOptions:
    """Options for the TDGL solver.

    Args:
        solve_time: Total simulation time, after any thermalization.
        skip_time: Amount of 'thermalization' time to simulate before recording data.
        dt_init: Initial time step.
        dt_max: Maximum adaptive time step.
        adaptive: Whether to use an adpative time step. Setting ``dt_init = dt_max``
            is equivalent to setting ``adaptive = False``.
        adaptive_window: Number of most recent solve steps to consider when
            computing the time step adaptively.
        max_solve_retries: The maximum number of times to reduce the time step in a
            given solve iteration before giving up.
        adaptive_time_step_multiplier: The factor by which to multiple the time
            step ``dt`` for each adaptive solve retry.
        field_units: The units for magnetic fields.
        current_units: The units for currents.
        output_file: Path to an HDF5 file in which to save the data.
            If the file name already exists, a unique name will be generated.
            If ``output_file`` is ``None``, the solver results will not be saved
            to disk.
        save_every: Save interval in units of solve steps.
        progress_interval: Minimum number of solve steps between progress bar updates.
        include_screening: Whether to include screening in the simulation.
        max_iterations_per_step: The maximum number of screening iterations per solve
            step.
        screening_tolerance: Relative tolerance for the induced vector potential, used
            to evaluate convergence of the screening calculation within a single time
            step.
        screening_step_size: Step size :math:`\\alpha` for Polyak's method.
        screening_step_drag: Drag parameter :math:`\\beta` for Polyak's method.
    """

    solve_time: float
    skip_time: float = 0.0
    dt_init: float = 1e-6
    dt_max: float = 1e-1
    adaptive: bool = True
    adaptive_window: int = 10
    max_solve_retries: int = 10
    adaptive_time_step_multiplier: float = 0.25
    save_every: int = 100
    progress_interval: int = 0
    field_units: str = "mT"
    current_units: str = "uA"
    output_file: Union[os.PathLike, None] = None
    include_screening: bool = False
    max_iterations_per_step: int = 1000
    screening_tolerance: float = 1e-3
    screening_step_size: float = 1.0
    screening_step_drag: float = 0.5

    def validate(self) -> None:
        if self.dt_init > self.dt_max:
            raise SolverOptionsError("dt_init must be less than or equal to dt_max.")
        if not (0 < self.adaptive_time_step_multiplier < 1):
            raise SolverOptionsError(
                "adaptive_time_step_multiplier must be in (0, 1)"
                f" (got {self.adaptive_time_step_multiplier})."
            )
        if not (0 < self.screening_step_drag <= 1):
            raise SolverOptionsError(
                "screening_step_drag must be in (0, 1)"
                f" (got {self.screening_step_drag})."
            )
        if self.screening_step_size <= 0:
            raise SolverOptionsError(
                "screening_step_size must be in > 0"
                f" (got {self.screening_step_size})."
            )
        if self.screening_tolerance <= 0:
            raise SolverOptionsError(
                "screening_tolerance must be in > 0"
                f" (got {self.screening_tolerance})."
            )



import warnings
from typing import Tuple, Union

import numpy as np
import scipy.sparse as sp

from .mesh import Mesh


def build_divergence(mesh: Mesh) -> sp.csr_matrix:
    """Build the divergence matrix that takes the divergence of a function living
    on the edges onto the sites.

    Args:
        mesh: The mesh.

    Returns:
        The divergence matrix.
    """
    edge_mesh = mesh.edge_mesh
    # Indices for each edge
    edge_indices = np.arange(len(edge_mesh.edges))
    # Compute the weights for each edge
    weights = edge_mesh.dual_edge_lengths
    # Rows and cols to update
    edges0 = edge_mesh.edges[:, 0]
    edges1 = edge_mesh.edges[:, 1]
    rows = np.concatenate([edges0, edges1])
    cols = np.concatenate([edge_indices, edge_indices])
    values = np.concatenate(
        [weights / mesh.areas[edges0], -weights / mesh.areas[edges1]]
    )
    return sp.csr_matrix(
        (values, (rows, cols)), shape=(len(mesh.sites), len(edge_mesh.edges))
    )


def build_gradient(
    mesh: Mesh,
    link_exponents: Union[np.ndarray, None] = None,
    weights: Union[np.ndarray, None] = None,
) -> sp.csr_matrix:
    """Build the gradient for a function living on the sites onto the edges.

    Args:
        mesh: The mesh.
        link_exponents: The value is integrated, exponentiated and used as
            a link variable.

    Returns:
        The gradient matrix.
    """
    edge_mesh = mesh.edge_mesh
    edge_indices = np.arange(len(edge_mesh.edges))
    if weights is None:
        weights = 1 / edge_mesh.edge_lengths
    if link_exponents is None:
        link_variable_weights = np.ones(len(weights))
    else:
        link_variable_weights = np.exp(
            -1j * np.einsum("ij, ij -> i", link_exponents, edge_mesh.directions)
        )
    rows = np.concatenate([edge_indices, edge_indices])
    cols = np.concatenate([edge_mesh.edges[:, 1], edge_mesh.edges[:, 0]])
    values = np.concatenate([link_variable_weights * weights, -weights])
    return sp.csr_matrix(
        (values, (rows, cols)), shape=(len(edge_mesh.edges), len(mesh.sites))
    )


def build_laplacian(
    mesh: Mesh,
    link_exponents: Union[np.ndarray, None] = None,
    fixed_sites: Union[np.ndarray, None] = None,
    free_rows: Union[np.ndarray, None] = None,
    fixed_sites_eigenvalues: float = 1,
    weights: Union[np.ndarray, None] = None,
) -> Tuple[sp.csc_matrix, np.ndarray]:
    """Build a Laplacian matrix on a given mesh.

    The default boundary condition is homogenous Neumann conditions. To get
    Dirichlet conditions, add fixed sites. To get non-homogenous Neumann condition,
    the flux needs to be specified using a Neumann boundary Laplacian matrix.

    Args:
        mesh: The mesh.
        link_exponents: The value is integrated, exponentiated and used as a
            link variable.
        fixed_sites: The sites to hold fixed.
        fixed_sites_eigenvalues: The eigenvalues for the fixed sites.

    Returns:
        The Laplacian matrix and indices of non-fixed rows.
    """
    if fixed_sites is None:
        fixed_sites = np.array([], dtype=int)

    edge_mesh = mesh.edge_mesh
    if weights is None:
        weights = edge_mesh.dual_edge_lengths / edge_mesh.edge_lengths
    if link_exponents is None:
        link_variable_weights = np.ones(len(weights))
    else:
        link_variable_weights = np.exp(
            -1j * np.einsum("ij, ij -> i", link_exponents, edge_mesh.directions)
        )
    edges0 = edge_mesh.edges[:, 0]
    edges1 = edge_mesh.edges[:, 1]
    rows = np.concatenate([edges0, edges1, edges0, edges1])
    cols = np.concatenate([edges1, edges0, edges0, edges1])
    areas0 = mesh.areas[edges0]
    areas1 = mesh.areas[edges1]
    values = np.concatenate(
        [
            weights * link_variable_weights / areas0,
            weights * link_variable_weights.conjugate() / areas1,
            -weights / areas0,
            -weights / areas1,
        ]
    )
    # Exclude all edges connected to fixed sites and set the
    # fixed site diagonal elements separately.
    if free_rows is None:
        free_rows = np.isin(rows, fixed_sites, invert=True)
    rows = rows[free_rows]
    cols = cols[free_rows]
    values = values[free_rows]
    rows = np.concatenate([rows, fixed_sites])
    cols = np.concatenate([cols, fixed_sites])
    values = np.concatenate(
        [values, fixed_sites_eigenvalues * np.ones(len(fixed_sites))]
    )
    laplacian = sp.csc_matrix(
        (values, (rows, cols)), shape=(len(mesh.sites), len(mesh.sites))
    )
    return laplacian, free_rows


def build_neumann_boundary_laplacian(
    mesh: Mesh, fixed_sites: Union[np.ndarray, None] = None
) -> sp.csr_matrix:
    """Build extra matrix for the Laplacian to set non-homogenous Neumann
    boundary conditions.

    Args:
        mesh: The mesh.
        fixed_sites: The fixed sites.

    Returns:
        The Neumann boundary matrix.
    """

    edge_mesh = mesh.edge_mesh
    boundary_index = np.arange(len(edge_mesh.boundary_edge_indices))
    # Get the boundary edges which are stored in the beginning of
    # the edge vector
    boundary_edges = edge_mesh.edges[edge_mesh.boundary_edge_indices]
    boundary_edges_length = edge_mesh.edge_lengths[edge_mesh.boundary_edge_indices]
    # Rows and cols to update
    rows = np.concatenate([boundary_edges[:, 0], boundary_edges[:, 1]])
    cols = np.concatenate([boundary_index, boundary_index])
    # The values
    values = np.concatenate(
        [
            boundary_edges_length / (2 * mesh.areas[boundary_edges[:, 0]]),
            boundary_edges_length / (2 * mesh.areas[boundary_edges[:, 1]]),
        ]
    )
    # Build the matrix
    neumann_laplacian = sp.csr_matrix(
        (values, (rows, cols)), shape=(len(mesh.sites), len(boundary_index))
    )
    # Change the rows corresponding to fixed sites to identity
    if fixed_sites is not None:
        # Convert laplacian to list of lists
        # This makes it quick to do slices
        neumann_laplacian = neumann_laplacian.tolil()
        # Change the rows corresponding to the fixed sites
        neumann_laplacian[fixed_sites, :] = 0

    return neumann_laplacian.tocsr()


class MeshOperators:
    """A container for the finite volume operators for a given mesh.

    Args:
        mesh: The :class:`tdgl.finite_volume.Mesh` instance for which to construct
            operators.
        fixed_sites: The indices of any sites for which the value of :math:`\\psi`
            and :math:`\\mu` are fixed as boundary conditions.
    """

    def __init__(self, mesh: Mesh, fixed_sites: Union[np.ndarray, None] = None):
        self.mesh = mesh
        edge_mesh = mesh.edge_mesh
        self.fixed_sites = fixed_sites
        self.laplacian_free_rows: Union[np.ndarray, None] = None
        self.divergence: Union[sp.spmatrix, None] = None
        self.mu_laplacian: Union[sp.spmatrix, None] = None
        self.mu_boundary_laplacian: Union[sp.spmatrix, None] = None
        self.mu_laplacian_lu: Union[sp.linalg.SuperLU, None] = None
        self.psi_gradient: Union[sp.spmatrix, None] = None
        self.psi_laplacian: Union[sp.spmatrix, None] = None
        self.link_exponents: Union[np.ndarray, None] = None
        # Compute these quantities just once, as they never change.
        self.gradient_weights = 1 / edge_mesh.edge_lengths
        self.laplacian_weights = edge_mesh.dual_edge_lengths / edge_mesh.edge_lengths
        self.gradient_link_rows = np.arange(len(edge_mesh.edges), dtype=int)
        self.gradient_link_cols = edge_mesh.edges[:, 1]
        self.laplacian_link_rows = np.concatenate(
            [edge_mesh.edges[:, 0], edge_mesh.edges[:, 1]]
        )
        self.laplacian_link_cols = np.concatenate(
            [edge_mesh.edges[:, 1], edge_mesh.edges[:, 0]]
        )

    def build_operators(self) -> None:
        """Construct the vector potential-independent operators."""
        mesh = self.mesh
        self.mu_laplacian, _ = build_laplacian(mesh, weights=self.laplacian_weights)
        self.mu_laplacian_lu = sp.linalg.splu(self.mu_laplacian)
        self.mu_boundary_laplacian = build_neumann_boundary_laplacian(mesh)
        self.mu_gradient = build_gradient(mesh, weights=self.gradient_weights)
        self.divergence = build_divergence(mesh)

    def set_link_exponents(self, link_exponents: np.ndarray) -> None:
        """Set the link variables and construct the covarient gradient
        and Laplacian for psi.

        Args:
            link_exponents: The value is integrated, exponentiated and used as
                a link variable.
        """
        mesh = self.mesh
        if link_exponents is not None:
            link_exponents = np.asarray(link_exponents)
        self.link_exponents = link_exponents
        if self.psi_gradient is None:
            # Build the matrices from scratch
            self.psi_gradient = build_gradient(
                mesh,
                link_exponents=link_exponents,
                weights=self.gradient_weights,
            )
            self.psi_laplacian, self.laplacian_free_rows = build_laplacian(
                mesh,
                link_exponents=link_exponents,
                fixed_sites=self.fixed_sites,
                free_rows=self.laplacian_free_rows,
                weights=self.laplacian_weights,
            )
            return
        # Just update the link variables
        directions = mesh.edge_mesh.directions
        if self.link_exponents is None:
            link_variables = np.ones(len(directions))
        else:
            link_variables = np.exp(
                -1j * np.einsum("ij, ij -> i", self.link_exponents, directions)
            )
        with warnings.catch_warnings():
            # This is slightly faster than re-creating the sparse matrices
            # from scratch.
            warnings.filterwarnings("ignore", category=sp.SparseEfficiencyWarning)
            # Update gradient for psi
            values = self.gradient_weights * link_variables
            rows, cols = self.gradient_link_rows, self.gradient_link_cols
            self.psi_gradient[rows, cols] = values
            # Update Laplacian for psi
            areas0 = mesh.areas[mesh.edge_mesh.edges[:, 0]]
            areas1 = mesh.areas[mesh.edge_mesh.edges[:, 1]]
            # Only update rows that are not fixed by boundary conditions
            free_rows = self.laplacian_free_rows[: len(self.laplacian_link_rows)]
            rows = self.laplacian_link_rows[free_rows]
            cols = self.laplacian_link_cols[free_rows]
            values = np.concatenate(
                [
                    self.laplacian_weights * link_variables / areas0,
                    self.laplacian_weights * link_variables.conjugate() / areas1,
                ]
            )[free_rows]
            self.psi_laplacian[rows, cols] = values



import argparse
import logging

from .visualization import (
    DEFAULT_QUANTITIES,
    InteractivePlot,
    MultiInteractivePlot,
    Quantity,
    create_animation,
)

logger = logging.getLogger("visualize")


def make_parser():
    parser = argparse.ArgumentParser(description="Visualize TDGL simulation data.")
    subparsers = parser.add_subparsers()
    parser.add_argument("--input", type=str, help="H5 file to visualize.")
    parser.add_argument(
        "-v",
        "--verbose",
        action="store_true",
        default=False,
        help="Run in verbose mode.",
    )
    parser.add_argument(
        "-s",
        "--silent",
        action="store_true",
        default=False,
        help="Run in silent mode.",
    )

    interactive_parser = subparsers.add_parser(
        "interactive", help="Create an interactive plot of one or more quantities."
    )
    interactive_parser.add_argument(
        "-q",
        "--quantities",
        type=lambda s: str(s).upper(),
        choices=Quantity.get_keys() + ["ALL"],
        nargs="*",
        help=(
            "Name(s) of the quantities to show. Because ``quantities`` takes a "
            "variable number of arguments, it must be the last argument provided."
        ),
    )

    interactive_parser.set_defaults(func=visualize_tdgl)

    animate_parser = subparsers.add_parser(
        "animate", help="Create an animation of the TDGL data."
    )
    animate_parser.add_argument(
        "-o", "--output", type=str, help="Output file for animation."
    )
    animate_parser.add_argument(
        "-f", "--fps", type=int, default=30, help="Frame rate of the animation."
    )
    animate_parser.add_argument(
        "-d", "--dpi", type=float, default=200, help="Resolution in dots per inch."
    )
    animate_parser.add_argument(
        "--min-frame",
        type=int,
        default=0,
        help="The first frame to render.",
    )
    animate_parser.add_argument(
        "--max-frame",
        type=int,
        default=-1,
        help="The last frame to render (-1 indicates the last step in the simulation).",
    )
    animate_parser.add_argument(
        "--axes-off",
        action="store_true",
        help="Turn the axes off.",
    )
    animate_parser.add_argument(
        "--title-off",
        action="store_true",
        help="Turn figure title off.",
    )
    animate_parser.add_argument(
        "-q",
        "--quantities",
        type=lambda s: str(s).upper(),
        choices=Quantity.get_keys() + ["ALL"],
        nargs="*",
        help=(
            "Name(s) of the quantities to show. Because 11quantities11 takes a "
            "variable number of arguments, it must be the last argument provided."
        ),
    )

    animate_parser.set_defaults(func=animate_tdgl)

    return parser


def animate_tdgl(args):
    kwargs = dict(
        input_file=args.input,
        output_file=args.output,
        logger=logger,
        silent=args.silent,
        dpi=args.dpi,
        fps=args.fps,
        min_frame=args.min_frame,
        max_frame=args.max_frame,
        axes_off=args.axes_off,
        title_off=args.title_off,
    )
    if args.quantities is None or "ALL" in args.quantities:
        kwargs["quantities"] = DEFAULT_QUANTITIES
    else:
        kwargs["quantities"] = args.quantities
    create_animation(**kwargs)


def visualize_tdgl(args):
    if args.quantities is None:
        InteractivePlot(input_file=args.input, logger=logger).show()
        return
    kwargs = dict(input_file=args.input, logger=logger)
    if "ALL" not in args.quantities:
        kwargs["quantities"] = args.quantities
    MultiInteractivePlot(**kwargs).show()


def main(args):
    logger.setLevel(logging.DEBUG if args.verbose else logging.INFO)
    logger.disabled = args.silent
    args.func(args)


if __name__ == "__main__":
    main(make_parser().parse_args())



import logging
from typing import List, Tuple, Union

import numpy as np
from meshpy import triangle
from scipy import spatial
from shapely.geometry.polygon import Polygon

from ..finite_volume.util import get_edge_lengths
from ..geometry import ensure_unique

logger = logging.getLogger(__name__)


def generate_mesh(
    poly_coords: np.ndarray,
    hole_coords: Union[List[np.ndarray], None] = None,
    min_points: Union[int, None] = None,
    max_edge_length: Union[float, None] = None,
    convex_hull: bool = False,
    boundary: Union[np.ndarray, None] = None,
    min_angle: float = 32.5,
    **kwargs,
) -> Tuple[np.ndarray, np.ndarray]:
    """Generates a Delaunay mesh for a given set of polygon vertex coordinates.

    Additional keyword arguments are passed to ``triangle.build()``.

    Args:
        poly_coords: Shape ``(n, 2)`` array of polygon ``(x, y)`` coordinates.
        hole_coords: A list of arrays of hole boundary coordinates.
        min_points: The minimimum number of vertices in the resulting mesh.
        max_edge_length: The maximum distance between vertices in the resulting mesh.
        convex_hull: If True, then the entire convex hull of the polygon (minus holes)
            will be meshed. Otherwise, only the polygon interior is meshed.
        boundary: Shape ``(m, 2)`` (where ``m <= n``) array of ``(x, y)`` coordinates
            for points on the boundary of the polygon.
        min_angle: The minimum angle in the mesh's triangles. Setting a larger value
            will make the triangles closer to equilateral, but the mesh generation
            may fail if the value is too large.

    Returns:
        Mesh vertex coordinates and triangle indices.
    """
    poly_coords = ensure_unique(poly_coords)
    if hole_coords is None:
        hole_coords = []
    hole_coords = [ensure_unique(coords) for coords in hole_coords]
    # Facets is a shape (m, 2) array of edge indices.
    # coords[facets] is a shape (m, 2, 2) array of edge coordinates:
    # [(x0, y0), (x1, y1)]
    coords = np.concatenate([poly_coords] + hole_coords, axis=0)
    xmin = coords[:, 0].min()
    dx = np.ptp(coords[:, 0])
    ymin = coords[:, 1].min()
    dy = np.ptp(coords[:, 1])
    r0 = np.array([[xmin, ymin]]) + np.array([[dx, dy]]) / 2
    # Center the coordinates at (0, 0) to avoid floating point issues.
    coords = coords - r0
    indices = np.arange(len(poly_coords), dtype=int)
    if convex_hull:
        if boundary is not None:
            raise ValueError(
                "Cannot have both boundary is not None and convex_hull = True."
            )
        facets = spatial.ConvexHull(coords).simplices
    else:
        if boundary is not None:
            boundary = list(map(tuple, boundary))
            indices = [i for i in indices if tuple(coords[i]) in boundary - r0]
        facets = np.array([indices, np.roll(indices, -1)]).T
    # Create facets for the holes.
    for hole in hole_coords:
        hole_indices = np.arange(
            indices[-1] + 1, indices[-1] + 1 + len(hole), dtype=int
        )
        hole_facets = np.array([hole_indices, np.roll(hole_indices, -1)]).T
        indices = np.concatenate([indices, hole_indices], axis=0)
        facets = np.concatenate([facets, hole_facets], axis=0)

    mesh_info = triangle.MeshInfo()
    mesh_info.set_points(coords)
    mesh_info.set_facets(facets)
    if hole_coords:
        # Triangle allows you to set holes by specifying a single point
        # that lies in each hole. Here we use the centroid of the hole.
        holes = [
            np.array(Polygon(hole).centroid.coords[0]) - r0.squeeze()
            for hole in hole_coords
        ]
        mesh_info.set_holes(holes)

    if "min_angle" not in kwargs:
        kwargs["min_angle"] = min_angle

    mesh = triangle.build(mesh_info=mesh_info, **kwargs)
    points = np.array(mesh.points) + r0
    triangles = np.array(mesh.elements)
    if min_points is None and (max_edge_length is None or max_edge_length <= 0):
        return points, triangles

    kwargs = kwargs.copy()
    kwargs["max_volume"] = dx * dy / 100
    i = 1
    if min_points is None:
        min_points = 0
    if max_edge_length is None or max_edge_length <= 0:
        max_edge_length = np.inf
    max_length = get_edge_lengths(points, triangles).max()
    while (len(points) < min_points) or (max_length > max_edge_length):
        mesh = triangle.build(mesh_info=mesh_info, **kwargs)
        points = np.array(mesh.points) + r0
        triangles = np.array(mesh.elements)
        max_length = get_edge_lengths(points, triangles).max()
        logger.debug(
            f"Iteration {i}: Made mesh with {len(points)} points and "
            f"{len(triangles)} triangles with maximum edge length: "
            f"{max_length:.2e}. Target maximum edge length: {max_edge_length:.2e}."
        )
        if np.isfinite(max_edge_length):
            kwargs["max_volume"] *= min(0.98, np.sqrt(max_edge_length / max_length))
        else:
            kwargs["max_volume"] *= 0.98
        i += 1
    return points, triangles
